/* ---------------------------------------------------------------- */
//
// OverView.m
//
// Created by Christoph Birk on 2011-12-08
// Copyright 2011 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#ifndef DEBUG 
#define DEBUG           1              // debug level
#endif

#define TIME_TEST       0

/* ---------------------------------------------------------------- */

#import  "OverView.h"
#import  "QltoolController.h"
#import  "Yutils.h"

/* ---------------------------------------------------------------- */

static void sample_image(int* rdata,int rw,int rh,int mx,int my,
                         int* sdata,int satlev)
{
  int *d;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d,%d,%d,%d,%p)\n",PREFUN,rdata,rw,rh,mx,my,sdata);
#endif

#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif
  if ((mx == 1) && (my == 1)) {
    memcpy(sdata,rdata,rw*rh*sizeof(int));
  } else { int img_w=rw/mx,img_h=rh/my,x,y,x1,y1,y2,a,n,satflag,v;
    for (y=0; y<img_h; y++) {            // create averaged image
      for (x=0,d=rdata+y*my*rw; x<img_w; x++,d+=mx) {
        satflag = a = n = 0;
        for (y1=0,y2=0; y1<my; y1++,y2+=rw) {
          for (x1=0; x1<mx; x1++) {
            v = (int)d[x1+y2];
            if (v >= satlev) { satflag = 1; break; } // any pixel saturated
            if (v == 0) continue;
            a += v; n++;
          }
          if (satflag) break;
        }
        if      (satflag) *sdata++ = satlev;
        else if (n == 0)  *sdata++ = 0;
        else              *sdata++ = a/n;
      } // endfor(x)
    } // endfor(y)
  } // endif(mx=my=1)
#if (TIME_TEST > 1)
  printf("%s(): cpu=%.1f [ms]\n",PREFUN,YmsSince(date));
#endif
}

/* ---------------------------------------------------------------- */

static void create_pixels(int* sdata,int w,int h,int mi,int dy,
                          u_char* pdata,int satlev)
{
  int i,n=w*h,c;
  
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif
  for (i=0; i<n; i++,sdata++,pdata++) {      // create pixel image
    if      (*sdata >= satlev) *pdata = 0xff;
    else if (*sdata == 0     ) *pdata = 0;
    else {
      c = (QLT_NCOLS*(*sdata - mi))/dy;
      if (c >= QLT_NCOLS) *pdata = QLT_NCOLS-1;
      else if (c < 0)     *pdata = 0;
      else                *pdata = c;
    }
  } // endfor(i)
#if (TIME_TEST > 1)
  printf("%s(): cpu=%.1f [ms]\n",PREFUN,YmsSince(date));
#endif
}

/* ---------------------------------------------------------------- */

@implementation OverView

@synthesize controller;

/* ---------------------------------------------------------------- */
#pragma mark "NSImageView"
/* ---------------------------------------------------------------- */

- (id)initWithFrame:(NSRect)frame
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.0f,%.0f)\n",PREFUN,frame.size.width,frame.size.height);
#endif
  
  self = [super initWithFrame:frame];
  if (self) { 
    int IM_W=frame.size.width,IM_H=(int)frame.size.height,j; 
    u_int *idata,*d,empty_color=0x00f0a0a0; // 0BGR
    d = idata = (u_int*)malloc(IM_W*IM_H*sizeof(u_int)); 
    for (j=0; j<IM_W*IM_H; j++) *d++ = empty_color;
    YimageData(self,idata,IM_W,IM_H); 
    free((void*)idata);
    [self setImageScaling:NSImageScaleAxesIndependently]; // scaleToFit 
    [self setImageAlignment:NSImageAlignCenter];
#if 0 // enable tracking
    frame.origin.x = frame.origin.y = 0;
    NSTrackingArea *area = [[NSTrackingArea alloc] initWithRect:frame
                             options:NSTrackingMouseEnteredAndExited|
                                     NSTrackingActiveInKeyWindow|
                                     NSTrackingEnabledDuringMouseDrag
                             owner:self userInfo:nil];
    [self addTrackingArea:area];
#endif
    safeBounds = [self bounds];
  }
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc                    
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  if (image_samp) { free((void*)image_samp); image_samp = NULL; }
  if (image_pix) { free((void*)image_pix); image_pix = NULL; }
  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (void)drawRect:(NSRect)rect
{
  CGContextRef context;                // NSGraphicsContext is not MT-safe
  float        x,y,w,h,mult_x,mult_y;
  QltoolController *qltc=(QltoolController*)controller;
#if (DEBUG > 1) 
  fprintf(stderr,"%s(%d,[%.0f,%.0f])\n",PREFUN,chip,rect.size.width,rect.size.height);
#endif
  assert([NSThread isMainThread]);     // NSView is mainThread only

  [super drawRect:rect];               // draw image
  safeBounds = [self bounds];

  if (!controller) return;  
  if (![self lockFocusIfCanDraw]) return;

  context = [[NSGraphicsContext currentContext] graphicsPort];
  mult_x = (qltc.raw_w / self.bounds.size.width);
  mult_y = (qltc.raw_h / self.bounds.size.height);
  
  NSRect frame = [qltc.mag_view frame];
  x =  qltc.mag_x/mult_x;
  y = (qltc.raw_h-qltc.mag_y)/mult_y;
  w = (frame.size.width/qltc.mag)/mult_x;
  h = (frame.size.height/qltc.mag)/mult_y;
  switch (qltc.lut) {
  case LUT_RAIN:
    CGContextSetRGBStrokeColor(context,0.0f,0.0f,0.0f,1); // black
    break;
  default:
    CGContextSetRGBStrokeColor(context,0.0f,1.0f,0.0f,1); // green
    break;
  }
  CGContextSetLineWidth(context,1);
  CGContextStrokeRect(context,CGRectMake(x-w/2.0f,y-h/2.0,w,h));
  
  [self unlockFocus];
}

/* ---------------------------------------------------------------- */

- (void)keyDown:(NSEvent*)event
{
  int dx=0,dy=0;
  QltoolController *qltc=(QltoolController*)controller;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p), keyCode=%d\n",PREFUN,event,[event keyCode]);
#endif

  NSUInteger mod = [event modifierFlags];

  NSString *chars = [event charactersIgnoringModifiers];
  switch ([chars characterAtIndex:0]) {
    case NSLeftArrowFunctionKey:  dx = -1; dy =  0; break;
    case NSRightArrowFunctionKey: dx = +1; dy =  0; break;
    case NSDownArrowFunctionKey:  dx =  0; dy = +1; break;
    case NSUpArrowFunctionKey:    dx =  0; dy = -1; break;
  }  
  if ((dx == 0) && (dy == 0)) {     // not cursor
    if ([chars compare:@" "] == NSOrderedSame) {   // FWHM fit
      [qltc update_fwhm:YES];
    }
    return;
  }

  if (mod & NSEventModifierFlagControl)   { dx *= 5; dy *=5; } 
  if (mod & NSEventModifierFlagOption)    { dx *= 5; dy *=5; }
  if (mod & NSEventModifierFlagCommand)   { dx *= 5; dy *=5; }
  if (mod & NSEventModifierFlagShift)     { dx *= 5; dy *=5; }
  [qltc incrementMouse:dx dy:dy];
  qltc.fwhm = 0;               // clear FWHM stats
  [qltc update_magnifier];

#if 0 // do not warp pointer
  safeBounds = [self bounds];
  float mult_x = (qltc.raw_w / bounds.size.width);
  float mult_y = (qltc.raw_h / bounds.size.height);
  NSPoint v; v.x =  qltc.mag_x / mult_x;
  v.y = (qltc.raw_h - qltc.mag_y) / mult_y;
  NSPoint p = [self convertPoint:v toView:nil];
  NSRect rect = [[self window] frame];
  CGPoint c; c.x = p.x + rect.origin.x;
  c.y = CGDisplayPixelsHigh(kCGDirectMainDisplay) - (p.y + rect.origin.y + 1);
  CGWarpMouseCursorPosition(c);
#endif
}

/* ---------------------------------------------------------------- */

- (void)mouseDown:(NSEvent*)event
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,chip,event);
#endif

  NSEventType typ = [event type];
  NSUInteger mod = [event modifierFlags];

  if ((typ == NSEventTypeRightMouseDown) || (mod & NSEventModifierFlagControl)) {
    [self squeeze_colormap:event]; 
  } else
  if (typ == NSEventTypeLeftMouseDown) {
    [self place_magnifier:event];
  }
  [[self window] makeFirstResponder:self];
}

/* ---------------------------------------------------------------- */

- (void)rightMouseDown:(NSEvent*)event
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,chip,event);
#endif
  [self squeeze_colormap:event]; 
}

/* ---------------------------------------------------------------- */

- (void)mouseDragged:(NSEvent*)event
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p)\n",PREFUN,event);
#endif
  assert([NSThread isMainThread]);     // NSView is mainThread only

  NSUInteger mod = [event modifierFlags];

  if (mod & NSEventModifierFlagControl) { 
    [self squeeze_colormap:event];
  } else 
    if (!(mod & (NSEventModifierFlagOption | NSEventModifierFlagCommand | NSEventModifierFlagShift))) {
    NSPoint loc = [self convertPoint:[event locationInWindow] fromView:nil];
    BOOL onView = [self mouse:loc inRect:[self bounds]];
    if (onView) [self place_magnifier:event];
  }
}

/* ---------------------------------------------------------------- */

- (void)rightMouseDragged:(NSEvent*)event
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,chip,event);
#endif
  [self squeeze_colormap:event];
}

/* ---------------------------------------------------------------- */

#if 0 // NSTrackingArea
//- (void)mouseExited:(NSEvent*)event
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,chip,event);
#endif
}

/* --- */

//- (void)mouseEntered:(NSEvent*)event
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,chip,event);
#endif
}
#endif

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */

- (void)newData:(int*)data min:(int)mi dyn:(int)dy
{
  int mult_x,mult_y,raw_w,raw_h,img_n;
  QltoolController *qltc=(QltoolController*)controller;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d,%d)\n",PREFUN,data,mi,dy);
#endif
  assert(controller);
#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif

  qltc.fwhm = 0;
  
  raw_w  = qltc.raw_w;
  raw_h  = qltc.raw_h;
  mult_x = (int)fmax(1,floor(raw_w/safeBounds.size.width)/2);
  mult_y = (int)fmax(1,floor(raw_h/safeBounds.size.height)/2);
  img_w  = raw_w/mult_x;
  img_h  = raw_h/mult_y;
  img_n  = img_w * img_h;
  image_samp = (int*)realloc(image_samp,img_n*sizeof(int));
  sample_image(data,raw_w,raw_h,mult_x,mult_y,image_samp,qltc.satlev);
  assert(QLT_NCOLS < 0xfe); 
  image_pix = (u_char*)realloc(image_pix,img_n*sizeof(u_char));
  create_pixels(image_samp,img_w,img_h,mi,dy,image_pix,qltc.satlev);
  [self newPixels];
#if (TIME_TEST > 0)
  printf("%s(): cpu=%4.1f [ms]\n",PREFUN,YmsSince(date));
#endif

  if (qltc.fwhm > 0) [qltc update_fwhm:YES];
}

/* ---------------------------------------------------------------- */

- (void)newPixels
{
  int     img_n=img_w*img_h,i;
  u_int   *image_data,*pixels;

#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif
  assert(controller);

  if (!image_pix) return;
  pixels = [controller pixels]; 

  assert(sizeof(u_int) == 4);
  image_data = (u_int*)malloc(img_n*sizeof(u_int));
  for (i=0; i<img_n; i++) image_data[i] = pixels[image_pix[i]];
  YimageData(self,image_data,img_w,img_h); // mainThread
  free((void*)image_data);
  
#if (TIME_TEST > 0)
  printf("%s(): cpu=%4.1f [ms]\n",PREFUN,YmsSince(date));
#endif
}

/* ---------------------------------------------------------------- */

- (void)squeeze_colormap:(NSEvent*)event
{
  double           contrast,level;
  QltoolController *qltc=(QltoolController*)controller;
#if (DEBUG > 1) 
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,chip,event);
#endif
  assert(controller);

  if (event) { NSRect frame; NSPoint loc; float x,y;
    frame = [[self window] frame]; 
    loc    = [event locationInWindow];
    x      = loc.x + 40.0f;
    y      = frame.size.height - loc.y -10.0f;
    contrast = exp((frame.size.height/2.0-y)/(frame.size.height/6.0));
    level    = (frame.size.width/2.0-x)/(frame.size.width/2.0);
  } else {
    contrast = 1.0f; level = 0.0f;
  }
  qltc.contrast = contrast;
  qltc.slevel = level;
  [qltc fill_pixels];
  if (event) [qltc redrawPixels];
}

/* ---------------------------------------------------------------- */

- (NSPoint)getMouse:(NSEvent*)event
{
  QltoolController *qltc=(QltoolController*)controller;
  assert(event); // NOTE: locationInWindow returns 'y' starting at '1'
  
  NSPoint loc = [self convertPoint:[event locationInWindow] fromView:nil];
  double mult_x = (qltc.raw_w / self.bounds.size.width);
  double mult_y = (qltc.raw_h / self.bounds.size.height);
  double x = loc.x * mult_x;
  double y = qltc.raw_h - (loc.y-1) * mult_y;
  
  return NSMakePoint(x,y);
}

/* ---------------------------------------------------------------- */

- (void)place_magnifier:(NSEvent*)event
{
  QltoolController *qltc=(QltoolController*)controller;
#if (DEBUG > 1) 
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,chip,event);
#endif
  assert(qltc);
  
  [qltc setMouse:[self getMouse:event]];
  qltc.fwhm = 0;
  [qltc update_magnifier];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
