/* ---------------------------------------------------------------- */
//
// QltoolController.m
//
// Created by Christoph Birk on 2011-12-08
// Copyright 2011 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           1              // debug level
#endif

#define TIME_TEST       0
#define GFIT_TEST       0

/* ---------------------------------------------------------------- */

#import  "QltoolController.h"
#import  "main.h"

/* ---------------------------------------------------------------- */

#define DBE_FRAME_QLT   "FRAME_QLT_Overview"
#define DBE_FRAME_QLM   "FRAME_QLT_Magnifier"

#define DARN_BIG   0x7fff0000
#define MAX_BYTE   0xff

/* ---------------------------------------------------------------- */

static const u_int satcol=0x000000ff;      // 0bgR

/* ---------------------------------------------------------------- */

static int linear(int value,double x1,double x2,int ncolors)
{
  double  x;
 
  x = (double)value/(double)ncolors;
  x = (x-x1)/(x2-x1);
  if (x < 0.0) x = 0.0;
  if (x > 1.0) x = 1.0;
 
  return (int)(x*MAX_BYTE);
}

/* ---------------------------------------------------------------- */
 
static int circle(int value,double m,double radius,int virtcols)
{
  double x;
 
  x = (double)value/(double)(virtcols-1);
 
  if ((x-m)*(x-m) >= radius*radius) return 0;
  
  return (int)(MAX_BYTE*sqrt(radius*radius - (x-m)*(x-m))/radius);
}

/* ---------------------------------------------------------------- */

static void i_sort(int *zahl, int l, int r)
{
  int  i,j,x,h;
 
  i = l;
  j = r;
 
  if (j>i) {
    x = zahl[(i+j)/2];
    do {
      while (zahl[i] < x) i++;
      while (zahl[j] > x) j--;
      if (i<=j) {
        h         = zahl[i];
        zahl[i++] = zahl[j];
        zahl[j--] = h;
      }
    } while (!(i>j));
    i_sort(zahl,l,j);
    i_sort(zahl,i,r);
  }
}

/* ---------------------------------------------------------------- */

static void scale_minmax(int* data,int w,int h,int* mi,int* ma,int sl)
{
  register int i,npix=w*h;
#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif

  *mi =  DARN_BIG;
  *ma = -DARN_BIG;
  for (i=0; i<npix; i++, data++) {
    if ((*data == 0) || (*data > sl)) continue;
    if ((int)*data > *ma) *ma = (int)*data;
    if ((int)*data < *mi) *mi = (int)*data;
  }
#if (TIME_TEST > 0)
  printf("%s(): cpu=%4.1f [ms] (%d,%d)\n",PREFUN,YmsSince(date),*mi,*ma);
#endif
}

/* ---------------------------------------------------------------- */

#if (GFIT_TEST > 0)
static double fn_gauss(double x,double a,double b)
{
  return(a * exp(-(x*x/(2.0*b*b))));
}
#endif

/* ---------------------------------------------------------------- */

static double get_background(int * data,int dimx,int dimy,
                             int x0,int y0,int r,double* noise)
{
  int     r2=r*r,n=0,x,y;
  int     *sortval,d;
  double  s1=0.0,s2=0.0,back=-1.0;

  sortval = (int*)malloc(4*r2*sizeof(int));

  for (x=x0-r; x<=x0+r; x++) {         /* find background */
    for (y=y0-r; y<=y0+r; y++) {       /* circular aperture */
      if ((x<0) || (x>=dimx) || (y<0) || (y>=dimy)) continue;
      if ((x-x0)*(x-x0)+(y-y0)*(y-y0) < r2) continue; /* use outside */
      d = data[x+y*dimx];
      if (d > 0) {
        sortval[n] = d; n++;
        if (noise) { s1 += (double)d; s2 += (double)d * (double)d; }
      }
    }
  }
  if (n > 0) {                         /* at least 1 pixel */
    i_sort(sortval,0,n-1);
    back = (double)sortval[n/2];
    if (noise) *noise = sqrt(s2/n - (s1/n)*(s1/n));
  }
  free((void*)sortval);

#if (DEBUG > 2)
  fprintf(stderr,"%s(): x0=%d, y0=%d, back=%.0f (n=%d)\n",PREFUN,x0,y0,back,n);
#endif

  return (double)back;
}

/* ---------------------------------------------------------------- */

static double get_centroid(int* data,int dimx,int dimy,
                           int x0,int y0,int r,double back,double* cx,double* cy)
{
  int    r2=r*r,x,y;
  double m=0.0,h;

  if (back < 0) back = get_background(data,dimx,dimy,x0,y0,r,NULL);
  if (back < 0) return back;
  
  *cx = *cy = 0.0;
  for (x=x0-r; x<=x0+r; x++) {         /* find centeroid */
    for (y=y0-r; y<=y0+r; y++) {
      if ((x<0) || (x>=dimx) || (y<0) || (y>=dimy)) continue;
      if ((x-x0)*(x-x0)+(y-y0)*(y-y0) > r2) continue; /* use inside */
      h = (double)data[x+y*dimx] - back;
      if (h < 0.0) continue;           /* signal too low */
      *cx += h*x;
      *cy += h*y;
      m   += h;
    }
  }
  if (m <= 0.0) return -1.0;
  *cx /= m;                            /* normalize centroid */
  *cy /= m;

#if (DEBUG > 2)
  fprintf(stderr,"%s(): cx=%.1f, cy=%.1f (m=%.1f)\n",PREFUN,*cx,*cy,m);
#endif

  return back;
}

/* ---------------------------------------------------------------- */

static double sqd(double x)
{
  return x*x;
}

/* ---------------------------------------------------------------- */

/* modified from 'Numerical Recipes in C' */

static double lfit(double* x,double* y,double* e,int n,double* a,double* b)
{
  int    i;
  double delta,h;
  double s=0.0,sx=0.0,sy=0.0,sxx=0.0,sxy=0.0,chi2=0.0;

  for (i=0; i<n; i++) {
    h    = e[i]*e[i];                  
    assert(h > 0);                     
    s   += 1.0/h;                       /* Warning: assumes all e[i] > 0 */
    sx  += x[i]/h;
    sy  += y[i]/h;
    sxx += x[i]*x[i]/h;
    sxy += x[i]*y[i]/h;
  }
  delta = s*sxx - sx*sx;
  *a    = (sxx*sy-sx*sxy)/delta;       /* y = a + bx */
  *b    = (s*sxy-sx*sy)/delta;
#if (DEBUG > 2)
  fprintf(stderr,"%s(): a=%f, b=%f\n",PREFUN,*a,*b);
#endif

  for (i=0; i<n; i++) {                /* chi-square */
    chi2 += sqd((y[i] - *a  - (*b)*x[i]) / e[i]);
  }

  return chi2;
} 

/* ---------------------------------------------------------------- */

static double gfit(double* x,double* y,int n,double* a,double* b,
                   double egain,double enoise)
{
  int    i;
  double *xval,*yval,*eval,chi2;

  xval = (double*)malloc(n*sizeof(double));
  yval = (double*)malloc(n*sizeof(double));
  eval = (double*)malloc(n*sizeof(double));

  for (i=0; i<n; i++) {                /* 'linearize' data */
    xval[i] = x[i]*x[i];               
    assert(y[i] > 0);                  /* Warning: assumes all y[i] > 0 */
    yval[i] = log(y[i]);               /* eval ~ 1/weight */
    eval[i] = (xval[i]+1) * ((sqrt(y[i]*egain+enoise*enoise)/egain)/y[i]);
  }
#if (GFIT_TEST > 0)
  { FILE *fp;
    fp = fopen("/tmp/linear.dat","w");
    for (i=0; i<n; i++) fprintf(fp,"%f  %f  %f\n",xval[i],yval[i],eval[i]);
    fclose(fp);
  }
#endif

  chi2 = lfit(xval,yval,eval,n,a,b);   /* linear regression fit */
  *a = exp(*a);                        /* 'de-linearize' */
  *b = sqrt(-0.5/(*b));
#if (DEBUG > 1)
  fprintf(stderr,"%s(): a=%f, b=%f\n",PREFUN,*a,*b);
#endif
  
  free((void*)xval); free((void*)yval); free((void*)eval);

  return chi2;
}
 

/* ---------------------------------------------------------------- */

static double peak2flux(double peak,double fwhm)
{
  return(1.133*peak*fwhm*fwhm);
}

/* --- */

static double get_fwhm(int* data,int dimx,int dimy,int x0,int y0,
                       int r,float* flx)
{
  auto int     x,y,r2=r*r,n;
  auto double  *flux,*dist,back,nbck=0,fwhm=0,a,b,d,cx=0,cy=0;
  const double egain=1.0,enoise=1.0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d,%d,%d,%d,%d)\n",PREFUN,data,dimx,dimy,x0,y0,r);
#endif
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif
  
  back = get_background(data,dimx,dimy,x0,y0,r,&nbck);
  if (back < 0) return 0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(): back=%.0f, noise=%.1f\n",PREFUN,back,nbck);
#endif
  back = get_centroid(data,dimx,dimy,x0,y0,r,back,&cx,&cy);
  if (back < 0) return 0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(): cx=%.1f, cy=%.1f\n",PREFUN,cx,cy);
#endif

  flux = (double*)malloc(4*r2*sizeof(double));
  dist = (double*)malloc(4*r2*sizeof(double));
  n    = 0;
  for (x=x0-r; x<=x0+r; x++) {         /* get flux */
    for (y=y0-r; y<=y0+r; y++) {       /* circular aperture */
      if ((x<0) || (x>=dimx) || (y<0) || (y>=dimy)) continue;
      d = ((double)x-cx)*((double)x-cx)+((double)y-cy)*((double)y-cy);
      if (d > r2) continue;            /* use inside */
      flux[n] = (double)data[x+y*dimx] - back;
      if (flux[n] < 1+3.0*nbck) continue;  /* flux too low */
      dist[n] = sqrt(d);
      n++;
    }
  }
#if (DEBUG > 1)
  fprintf(stderr,"%s(): n=%d\n",PREFUN,n);
#endif

  if (n > 1) {
#if (DEBUG > 1)
    double chi2 = gfit(dist,flux,n,&a,&b,egain,enoise);
#else
    gfit(dist,flux,n,&a,&b,egain,enoise); // Clang
#endif
    if ((a > 0) && (b > 0)) { double f; // sanity check
      fwhm = 2.355*b;                   // 2*sqrt(-2*ln(0.5))
      f = peak2flux(a,fwhm);            // Warning: valid only if pixscale<<fwhm
#if (DEBUG > 1)
      fprintf(stderr,"%s(): fwhm=%.1f, flux=%f, chi=%f\n",PREFUN,fwhm,f,sqrt(chi2/n));
#endif
      if (flx) *flx = (float)f;
#if (GFIT_TEST > 0)
      { int i; double d; FILE *fp;
        fp = fopen("/tmp/pixel.dat","w");
        for (i=0; i<n; i++) fprintf(fp,"%f %f\n",dist[i],flux[i]);
        fclose(fp);
        fp = fopen("/tmp/gauss.dat","w");
        for (d=0.; d<(double)r; d+=.05) fprintf(fp,"%f %f\n",d,fn_gauss(d,a,b));
        fclose(fp);
      }
#endif
    }
  } 
  free((void*)flux); free((void*)dist);

#if (TIME_TEST > 1)
  printf("%s(): cpu=%.1f [ms]\n",PREFUN,YmsSince(date));
#endif

  return fwhm;
}

/* ---------------------------------------------------------------- */

static const u_int outcol=0x0000a000;  // 0bGr

static void set_col(u_int* data,int x0,int y0,int w,int m,u_int c)
{
  int x,y,x1=x0+m,y1=y0+m;
  
  for (y=y0; y<y1; y++) {
    for (x=x0; x<x1; x++) {
      data[x+y*w] = c;
    }
  }
}

/* ---------------------------------------------------------------- */

@implementation QltoolController

@synthesize raw_w,raw_h;
@synthesize contrast,slevel;
@synthesize fwhm,pixels;
@synthesize satlev,lut;
@synthesize mag_view,mag,mag_x,mag_y;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */


/* ---------------------------------------------------------------- */
#pragma mark "NSWindowController"
/* ---------------------------------------------------------------- */

- (id)initWithWindowNib:(NSString*)name 
{
  NSInteger i=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  if (!(self = [super initWithWindowNibName:name])) return nil;

  flip_x = flip_y = 0;                // no flips
  raw_w = raw_h = 2048;               // default geometry
#if (DEBUG > 0)
  printf("[%d]: (%d,%d) (%d,%d)\n",(int)i,flip_x,flip_y,raw_w,raw_h);
#endif
  raw_data = (int*)malloc(raw_w*raw_h*sizeof(int));

  data_dyn = 1;
  contrast = 1.0;
  slevel = 0.0;
  
  YwindowFrameSet([self window],DBE_FRAME_QLT);
#ifdef NDEBUG
  i = [[self window] styleMask];
  [[self window] setStyleMask:(i & ~NSWindowStyleMaskClosable)];
#endif

  mag_view.controller = self;
  mag_view.mag = 4;                   // 'IB' dependency

  YwindowFrameSet([mag_view window],DBE_FRAME_QLM);
#ifdef NDEBUG
  i = [[mag_view window] styleMask];
  [[mag_view window] setStyleMask:(i & ~NSWindowStyleMaskClosable)]; 
#endif

  return self;
}

/* ---------------------------------------------------------------- */

- (void)windowDidLoad
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  assert(root_view);
  NSRect frame = [root_view frame];
  frame.origin.x = frame.origin.y = 0;
  over_view = [[OverView alloc] initWithFrame:frame];
  [root_view addSubview:over_view];
  over_view.controller = self;

  [self reloadPreferences:NO];

  stretch = (int)[popup_stretch indexOfSelectedItem];
  lut = (int)[popup_lut indexOfSelectedItem];
  [self fill_pixels];
  
  radius = [step_radius intValue];
  [edit_radius setIntValue:radius];

  [self action_mag:nil];
  [self action_radius:nil];
}

/* ---------------------------------------------------------------- */

- (void)close
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  YwindowFrameSave([self window],DBE_FRAME_QLT);
  YwindowFrameSave([mag_view window],DBE_FRAME_QLM);
  
  [[mag_view window] close];
  [super close];                       // reduces retainCount
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  mag_view.controller = nil;
  
  over_view.controller = nil;
  [over_view release]; over_view = nil;
  if (raw_data) { free((void*)raw_data); raw_data = NULL; }
  if (pixels) { free((void*)pixels); pixels = NULL; }

  [super dealloc];
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */

- (void)reloadPreferences:(BOOL)redraw
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,(int)redraw);
#endif
  
  int s = YprefGetInt(DBE_SATLEV,DEF_SATLEV,use_shared);
  if (s != satlev) {
    satlev = s;
    if (redraw) [self redrawWithScaling:NO];
  }
}

/* ---------------------------------------------------------------- */

- (void)fill_pixels
{
  int     i,s,virt,offset,ncols=QLT_NCOLS;
  u_int   col;
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d)\n",PREFUN,lut);
#endif
  assert(sizeof(u_int) == 4);
  assert(ncols < 0xfe);
  
  if (!pixels) pixels = (u_int*)malloc(0x100*sizeof(u_int));

  offset = (int)((double)ncols/2.0*(1.0-slevel-fabs(contrast)));
  for (i=0; i<ncols; i++) {
    virt = (int)((i-offset)/fabs(contrast)); /* virtual color cell  */
    if (contrast < 0.0)  virt = ncols-virt;
    if (virt < 0)        virt = 0;
    if (virt >= ncols)   virt = ncols-1;
    switch (lut){
    case LUT_BBDY:                     /* black body */
      col  = linear(virt,0.00,0.50,ncols);       // R
      col += linear(virt,0.25,0.75,ncols) << 8;  // G
      col += linear(virt,0.50,1.00,ncols) << 16; // B
      break;
    case LUT_RAIN:                     /* rainbow */
      col  = circle(virt,1.,.55,ncols);          // R
      col += circle(virt,.5,.45,ncols) << 8;     // G
      col += circle(virt,0.,.65,ncols) << 16;    // B
      break;
    case LUT_GREY:                     /* grayscale */
    case LUT_INV:                      /* inverted grayscale */
    default:
      if (lut == LUT_INV) s = (int)(((ncols-virt)*MAX_BYTE)/ncols);
      else                s = (int)(MAX_BYTE-((ncols-virt)*MAX_BYTE)/ncols);
      col = s + (s << 8) + (s << 16);
      break;
    }
    pixels[i] = col; // printf("col[%3d]=%06x\n",i,pixels[i]);
  }
  pixels[0xff] = satcol;
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions"
/* ---------------------------------------------------------------- */

- (IBAction)action_lut:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  lut = (int)[sender indexOfSelectedItem];
  [self fill_pixels];
  [self redrawPixels];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_stretch:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  stretch = (int)[sender indexOfSelectedItem];
  [edit_min setEnabled:(stretch) ? NO : YES];
  [edit_max setEnabled:(stretch) ? NO : YES];
  [over_view squeeze_colormap:nil];  // reset contrast/level
  [self redrawWithScaling:YES];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_minmax:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  if (sender == edit_min) data_min = [sender intValue];
  if (sender == edit_max) data_max = [sender intValue];

  [self redrawWithScaling:YES];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_mag:(id)sender
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  Ystr2chr([popup_mag titleOfSelectedItem],buf,sizeof(buf));
  mag_view.mag = mag = atoi(buf);
  mag_view.radius = mag*radius;
  if (sender) { 
    [over_view setNeedsDisplay:YES];
    [self update_magnifier];
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_radius:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  radius = [step_radius intValue];
  [edit_radius setIntValue:radius];
  mag_view.radius = mag * radius;
  if (sender) {
    [self update_magnifier];
    if (fwhm > 0) [self update_fwhm:YES]; 
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Image/Redraw"
/* ---------------------------------------------------------------- */

- (void)reset:(BOOL)clear width:(int)w height:(int)h title:(NSString*)tstr
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d,%d): %d,%d\n",PREFUN,(int)clear,w,h,raw_w,raw_h);
#endif
  
  mag_x = (int)((float)mag_x * ((float)w/(float)raw_w)); // keep magnifier
  mag_y = (int)((float)mag_y * ((float)h/(float)raw_h)); // in place

  if ((raw_w != w) || (raw_h != h)) {
    raw_data = (int*)realloc(raw_data,w*h*sizeof(int));
    raw_w = w;
    raw_h = h;
  }
  if (clear) {
    memset(raw_data,0,w*h*sizeof(int));
  }
  if (tstr) { 
    YwindowTitle([self window],[tstr UTF8String]);
  } else {
    YwindowTitle([self window],"OverView");
  }
}

/* ---------------------------------------------------------------- */

- (void)updateData:(int*)ptr line:(int)y
{
#if (DEBUG > 3)
  fprintf(stderr,"%s(%d,%p,%d\n",PREFUN,chip,ptr,y);
#endif
  
  int y_here = (flip_y) ? raw_h-y-1 : y;
  if (flip_x) { register int *p; register int i;
    p = raw_data+(y_here+1)*raw_w-1;
    for (i=0; i<raw_w; i++) *p-- = *ptr++; 
  } else {
    memcpy(raw_data+y_here*raw_w,ptr,raw_w*sizeof(int));
  }
}

/* ---------------------------------------------------------------- */

- (void)redrawWithScaling:(BOOL)rescale
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,rescale);
#endif

  assert(sizeof(u_int) == 4);
  dispatch_async(GLOBAL_QUEUE,^(void) {
#if (TIME_TEST > 0)
    NSDate *date = [NSDate date];
#endif
    if (rescale) [self scale_data];
    assert(data_dyn != 0);
    [over_view newData:raw_data min:data_min dyn:data_dyn];
    [self update_magnifier];
#if (TIME_TEST > 0)
    fprintf(stderr,"redrawTime=%.3f\n",YmsSince(date));
#endif
  });
}

/* ---------------------------------------------------------------- */

- (void)redrawPixels
{
  [over_view newPixels];
  [self update_magnifier];
}

/* ---------------------------------------------------------------- */

- (u_int*)createMagDataW:(int)mag_w H:(int)mag_h
{
  int    xx,yy,colour,x,y;
  u_int  *image_data,c;
  
  assert(data_dyn != 0);
  int img_w = mag_w * mag;
  int img_h = mag_h * mag;
  assert(sizeof(u_int) == 4);
  image_data = (u_int*)malloc(img_w*img_h*sizeof(u_int));

  for (xx=0; xx<mag_w; xx++) {         // create image
    x = mag_x - mag_w/2 + xx;
    for (yy=0; yy<mag_h; yy++) {
      y = mag_y - mag_h/2 + yy;
      if ((x<0) || (x>=raw_w) || (y < 0) || (y >= raw_h)) {
        c = outcol;
      } else
      if (raw_data[x+y*raw_w] >= satlev) {
        c = pixels[0xff];
      } else {
        colour = (QLT_NCOLS*(raw_data[x+y*raw_w] - data_min))/data_dyn;
        if (colour >= QLT_NCOLS) c = pixels[QLT_NCOLS-1];
        else if (colour < 0)     c = pixels[0];
        else                     c = pixels[colour];
      }
      set_col(image_data,xx*mag,yy*mag,img_w,mag,c);
    }
  }
  return image_data;
}

/* ---------------------------------------------------------------- */

- (void)update_magnifier
{
  int     x,y,v,n=0;
  int     aper_min=0xffff,aper_max=0;
  double  aper_ave,aper_sig,s1=0.0,s2=0.0;
  char    buf[128];
#if (DEBUG > 2)
  fprintf(stderr,"%s:%p: mThread=%d\n",PREFUN,self,[NSThread isMainThread]);
#endif

  assert((mag_x >= 0) && (mag_x < raw_w));
  assert((mag_y >= 0) && (mag_y < raw_h));
  sprintf(buf,"%d,%d = %d",flip_x ? raw_w-mag_x : mag_x+1,
                           flip_y ? raw_h-mag_y : mag_y+1,
                           raw_data[mag_x+mag_y*raw_w]);
  YeditSet(edit_pixel,buf);            // mainThread
  const int r = radius;
  for (x=mag_x-r; x<=mag_x+r; x++) {   // statistics
    if ((x<0) || (x>=raw_w)) continue;
    for (y=mag_y-r; y<=mag_y+r; y++) {
      if ((y<0) || (y>=raw_h)) continue;
      v = raw_data[x+raw_w*y];
      if ((v == 0) || (v >= satlev)) continue;
      if (v > aper_max) aper_max = v;
      if (v < aper_min) aper_min = v;
      s1 += (double)v; s2 += (double)v * (double)v;
      n++;
    }
  }
  aper_ave = (n>0) ? s1/n : 0.0;
  aper_sig = (n>1) ? sqrt(s2/n-aper_ave*aper_ave) : 0.0;
  sprintf(buf,"%d, %d (n=%d)",aper_min,aper_max,n);
  YeditSet(edit_minmax,buf);          // mainThread
  sprintf(buf,"%.1f, %.2f",aper_ave,aper_sig);
  YeditSet(edit_avesig,buf);

  [self update_fwhm:NO];

#if (TIME_TEST > 1)
  NSDate *date1 = [NSDate date];
#endif
  NSRect bounds = mag_view.safeBounds;
  int mag_w = (int)bounds.size.width/mag;  if (mag_w > 1) mag_w -= ((mag_w+1)%2);
  int mag_h = (int)bounds.size.height/mag; if (mag_h > 1) mag_h -= ((mag_h+1)%2);
  assert((mag_w > 0) && (mag_h > 0));
  u_int *image_data = [self createMagDataW:mag_w H:mag_h];
  YimageData(mag_view,image_data,mag_w*mag,mag_h*mag);
  free((void*)image_data);
#if (TIME_TEST > 1)
  NSDate *date2 = [NSDate date];
  fprintf(stderr,"magTime=%.2f [ms]\n",YmsSince(date));
#endif
  YviewNeedsDisplay(over_view);
}

/* ---------------------------------------------------------------- */

- (void)update_fwhm:(BOOL)recalc
{
  char buf[128];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d)\n",PREFUN,recalc);
#endif

  if (recalc) {
    fwhm = (float)get_fwhm(raw_data,raw_w,raw_h,mag_x,mag_y,radius,&flux);
  }
  if (fwhm > 0.0f) sprintf(buf,"%.1f / %.0f",fwhm,flux);
  else              strcpy(buf,"n/a");
  YeditSet(edit_fwhm,buf);             // mainThread
}

/* ---------------------------------------------------------------- */

- (void)scale_medF1:(float)f1 F2:(float)f2
{
  auto     int    i,n=0,m,lim1,lim2,median;
  auto     int    n_samp,tpix=raw_w*raw_h,timeout;
  auto     int    *copy0;
  auto     double sum1=0.0,sum2=0.0,sigma=0.0;
  register int    *copy;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.1f,%.1f): mThread=%d\n",PREFUN,f1,f2,[NSThread isMainThread]);
#endif
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif

  n_samp = (int)sqrt(tpix);
  timeout = (int)fmin(128*n_samp,tpix);
  copy = copy0 = (int*)malloc(n_samp*sizeof(int));
  while (n < n_samp) {                 // take random sample
    if (--timeout < 0) break;          // just in calse all pixels are invalid
    *copy = raw_data[arc4random_uniform(tpix)];
    if (*copy > satlev) continue;      // ignore saturated pixels
    copy++; n++;
  }
  if (n == 0) {
    data_min = data_max = 0; free((void*)copy0);
    return;
  }

  if (n > 1) i_sort(copy0,0,n-1);      // sort pixels
  median = copy0[n/2];
#if (DEBUG > 1)
  fprintf(stderr,"%s(): median=%d (%d,%d)\n",PREFUN,median,n,timeout);
#endif

  if (copy0[0] == copy0[n-1]) {        // all pixels equal
    data_min = median-1;
    data_max = median+1;
    free((void*)copy0);
    return;
  }

  if (f1 < 1.0) {                      // use fraction of pixels
    data_min = (int)copy0[(int)((1.0-f1)*n/2.0)]-1;
    data_max = (int)copy0[(int)((f1+(1.0-f1)/2.0)*n)]+1;
  } else {
    lim1 = median - f1*(median-copy0[n/4]) -1;
    lim2 = (int)fmin(satlev,median+f2*(copy0[(3*n)/4]-median)+1);
#if (DEBUG > 1)
    fprintf(stderr,"%s(): lim1=%d, lim2=%d\n",PREFUN,lim1,lim2);
#endif
    for (i=0,m=0,copy=copy0; i<n; i++,copy++) {
      if ((int)*copy > lim1) {         // calc. std-dev.
        if ((int)*copy < lim2) {
          sum1 += (double)*copy;
          sum2 += (double)*copy * (double)*copy;
          m++;
        }
      }
    }
    if (m) sigma = sqrt(sum2/m - (sum1/m)*(sum1/m));
    else   sigma = 1.0;
#if (DEBUG > 1)
    fprintf(stderr,"%s(): sigma=%.1f\n",PREFUN,sigma);
#endif
    data_min = median - (int)(f1*sigma+1.0);
    data_max = median + (int)(f2*sigma+1.0);
  }
 
  free((void*)copy0);

#if (TIME_TEST > 1)
  fprintf(stderr,"scaleTime=%.2f [ms]\n",YmsSince(date));
#endif
}

/* ---------------------------------------------------------------- */

- (void)scale_data
{
  switch (stretch) { 
  case STRETCH_MED3:               /* median -3/+3 sigma */
    [self scale_medF1:3.0f F2:3.0f];
    break;
  case STRETCH_MED5:               /* median -5/+5 sigma */
    [self scale_medF1:5.0f F2:5.0f];
    break;
  case STRETCH_310:                /* median -3/+10 sigma */
    [self scale_medF1:3.0f F2:10.0f];
    break;
  case STRETCH_MIMA:               /* min/max scaling */
    scale_minmax(raw_data,raw_w,raw_h,&data_min,&data_max,satlev);
    break;
  case STRETCH_CUTS:               /* manual scaling */
  default:
    break;
  }
  data_dyn = data_max - data_min;
  if (data_dyn == 0) data_dyn = 1;
  dispatch_async(MAIN_QUEUE,^(void) { char buf[32];
    sprintf(buf,"%d",data_min); YeditSet(edit_min,buf);
    sprintf(buf,"%d",data_max); YeditSet(edit_max,buf); // avoid 1,000
    sprintf(buf,"min=%d max=%d dyn=%d",data_min,data_max,data_dyn);
    [main_logger append:buf];
  });
}

/* ---------------------------------------------------------------- */
#pragma mark "Mouse"
/* ---------------------------------------------------------------- */

- (void)setMouse:(NSPoint)mouse
{
  mag_x = (int)fmax(0,fmin(raw_w-1,mouse.x));
  mag_y = (int)fmax(0,fmin(raw_h-1,mouse.y));
}

/* --- */

- (void)incrementMouse:(int)dx dy:(int)dy
{
  mag_x = (int)fmax(0,fmin(raw_w-1,mag_x+dx));
  mag_y = (int)fmax(0,fmin(raw_h-1,mag_y+dy));
}

/* ---------------------------------------------------------------- */
#pragma mark "NSWindow delegates"
/* ---------------------------------------------------------------- */

- (void)windowDidResize:(NSNotification*)note
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p): %p\n",PREFUN,note,[note object]);
#endif

  if ([[note object] inLiveResize]) return;

  if ([note object] == [self window]) {
    NSRect frame = [root_view frame];
    frame.origin.x = frame.origin.y = 0;
    [over_view setFrame:frame];
  } else
  if ([note object] == [mag_view window]) {
    [self update_magnifier];
  }
}

/* --- */

- (void)windowDidEndLiveResize:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): object=%p\n",PREFUN,note,[note object]);
#endif
  assert([NSThread isMainThread]);

  [self performSelector:@selector(windowDidResize:) withObject:note
        afterDelay:0.25];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
