/* ----------------------------------------------------------------
 * 
 * utils.c
 * 
 * Christoph C. Birk (birk@obs.carnegiescience.edu)
 *
 * ---------------------------------------------------------------- */

#define IS_UTILS_C

/* DEFINEs -------------------------------------------------------- */

#define DEBUG              1           // debug level

/* INCLUDEs ------------------------------------------------------- */

#include <stdlib.h>                    // atoi()
#include <stdio.h>                     // sprintf(),fprintf()
#include <string.h>                    // strlen()
#include <strings.h>                   // strcasecmp()
#include <ctype.h>                     // isspace()
#include <math.h>                      // sin(),atof()
#include <time.h>                      // time(),ctime()
#include <assert.h>
#include <fcntl.h>                     // O_RDWR
#include <unistd.h>                    // lockf()

#include <sys/time.h>                  // gettimeofday()

#include "utils.h"

/* ---------------------------------------------------------------- */

double get_sidereal(time_t ut,double tellong)
{
  double st;
  const double STG0=24055.77;          // minimizes diff.to get_st() 2015-2035

  if (ut == 0) ut = cor_time(0);       // now

  st = (double)((int)(STG0+366.2422/365.2422*(double)ut) % 86400) * (24.0/86400.0);
  st = modulo(st+tellong/15.0,0.0,24.0);
  
  return(st*3600.0);                   // seconds
}

/* --- */

static double get_julian(time_t ut)
{ // http://scienceworld.wolfram.com/astronomy/JulianDate.html
  struct tm ts;
  
  gmtime_r(&ut,&ts);
  double y = 1900.0+ts.tm_year;
  double m = 1.0+ts.tm_mon;
  double t = (double)ts.tm_hour/24.0+(double)ts.tm_min/1440.0+(double)ts.tm_sec/86400.0;
  double jd = 367.0*y - floor(7.0*(y+floor((m+9.0)/12.0))/4.0)
              - floor(3.0*(floor((y+(m-9.0)/7.0)/100.0)+1.0)/4.0)
              + floor(275.0*m/9.0) + ts.tm_mday + 1721028.5 + t;

  return jd;
}

/* --- */

double get_st(time_t ut,double tellong)
{ // http://aa.usno.navy.mil/faq/docs/GAST.php
  double d = get_julian(ut) - 2451545.0;
  double gmst = 18.697374558 + 24.06570982441908*d;
  double st = 3600.0 * modulo(gmst+tellong/15.0,0.0,24.0);
  
  return st;
}

/* ---------------------------------------------------------------- */

double get_airmass(double lat,double dec,double ha)
{
  double cos_zd,x,airmass;
  const double SCALE=750.0;            // atmospheric scale height

  cos_zd = sin(lat/RADS) * sin(dec/RADS) +
           cos(lat/RADS) * cos(dec/RADS) * cos(15.0*ha/RADS);
  if ((cos_zd > 1.0) || (cos_zd < 0.0)) return (9.999);
  x = SCALE * cos_zd;
  airmass = sqrt(x*x + 2.0*SCALE + 1.0) - x;
  if (airmass > 9.999) airmass = 9.999;

  return airmass;
}

/* ---------------------------------------------------------------- */

double get_ha(time_t ut,double tellong,double alpha)
{
  double ha;

  ha = get_sidereal(ut,tellong)/3600.0 - alpha;

  return ha;
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

#ifndef __isleap
#define __isleap(y) (((y) % 4 == 0) && (((y) % 100 != 0) || ((y) % 400 == 0)))
#endif

double get_epoch(time_t when)
{
  int        yr;
  double     ep,doy,hr,dpy;
  struct tm  res;

  time_t ut = (when) ? when : cor_time(0);
  gmtime_r(&ut,&res);

  yr  = 1900 + (int)res.tm_year;
  dpy = (__isleap(yr)) ? 366.0 : 365.0;
  doy = (double)res.tm_yday;
  // verifyed assert(doy < dpy);
  hr  = (double)res.tm_hour;

  if (when == 0) {                     // round to day
    ep = (double)yr + my_round(doy+hr/24.0,0)/dpy;
  } else { 
    double mi  = (double)res.tm_min;
    double se  = (double)res.tm_sec;
    ep = (double)yr + doy/dpy + hr/(24.0*dpy) + mi/(1440.0*dpy) + se/(86400.0*dpy);
  }
  
  return ep;
}

/* ---------------------------------------------------------------- */

char* get_night(char* night,time_t when)
{
  struct tm *ltime,res;
  time_t     ut;

  if (when == 0) ut = cor_time(0);
  else           ut = when;
  ut   -= 9*3600;                       // until 9am
  ltime = localtime_r(&ut,&res);        // get local time

  strftime(night,16,"%d%b%Y",ltime);

  return night;
}

/* ---------------------------------------------------------------- */

int get_uts(time_t ut)
{
  struct tm *gtime,res;

  gtime = gmtime_r(&ut,&res);           // GMT
 
  return((int)gtime->tm_hour*3600+(int)gtime->tm_min*60+(int)gtime->tm_sec);
}

/* ---------------------------------------------------------------- */

time_t cor_time(long delta)
{
  return(time(NULL) - delta);
}

/* ---------------------------------------------------------------- */

double walltime(void)
{
  struct timeval tp;
  
  gettimeofday(&tp,NULL);
  
  return((double)(tp.tv_sec)+(double)(tp.tv_usec)/1.0e6);
}

/* ---------------------------------------------------------------- */

char* time_str(char* text,double t)
{
  int hr,mi,se;

  hr = (int)floor(t/3600.0);
  t -= hr*3600.0;
  mi = (int)floor(t/60.0);
  t -= mi*60.0;
  se = (int)floor(t+0.5);
  if (se >= 60) { se -= 60; mi += 1; }
  if (mi >= 60) { mi -= 60; hr += 1; }
  if (hr >= 24) { hr -= 24; }
  sprintf(text,"%2d:%02d:%02d",hr,mi,se);

  return text;
}

/* ---------------------------------------------------------------- */

char* alpha_str(char* text,double alpha)
{
  double hr,mi,se;

  hr = floor(alpha);
  mi = floor((alpha-hr)*60.);
  se = (alpha-hr-mi/60.)*3600.;
  if (se >= 59.95) { se -= 59.95; mi += 1.; }
  if (mi >= 60.00) { mi -= 60.00; hr += 1.; }
  if (hr >= 24.00) { hr -= 24.00; }
  sprintf(text," %02.0f:%02.0f:%04.1f",hr,mi,se);

  return text;
}

/* ---------------------------------------------------------------- */

char* delta_str(char* text,double delta)
{
  double dg,pd,mi,se;
  char   sign='+';

  pd = fabs(delta);
  dg = floor(pd);
  mi = floor((pd-dg)*60.);
  se = (pd-dg-mi/60.)*3600.;
  if (se >= 59.95) { se -= 59.95; mi += 1.0; }
  if (mi >= 60.00) { mi -= 60.00; dg += 1.0; }
  if (delta < 0.0) sign = '-';
  sprintf(text,"%c%02.0f:%02.0f:%04.1f",sign,dg,mi,se);

  return text;
}

/* ---------------------------------------------------------------- */

double d_alpha(const char* text)
{
  double h=0.0,m=0.0,s=0.0;
  
  if (!strchr(text,(int)':'))
    sscanf(text,"%lf %lf %lf",&h,&m,&s);
  else
    sscanf(text,"%lf:%lf:%lf",&h,&m,&s);
  
  return modulo(h+(m/60.0)+(s/3600.0),0.0,24.0);
}

/* ---------------------------------------------------------------- */

double d_delta(const char* text)
{
  double deg=0.0,dem=0.0,des=0.0,d;

  if (!strchr(text,(int)':'))
    sscanf(text,"%lf %lf %lf",&deg,&dem,&des);
  else
    sscanf(text,"%lf:%lf:%lf",&deg,&dem,&des);

  d = fabs(deg)+fabs(dem)/60.0+fabs(des)/3600.0;
  while (d >= 90.0) d -= 90.0;

  if ((deg<0.0) || (dem<0.0) || (des<0.0) || strchr(text,(int)'-')) {
    d = -d;
  }

  return d;
}

/* ---------------------------------------------------------------- */

#if 0 // unused
double d_equinox(const char* text)
{
  double e;

  e = atof(text);
  if (e <= 0.0) e = get_epoch(0);

  return e;
}
#endif

/* ---------------------------------------------------------------- */


void precess(double ain,double din,double ein, 
             double *aout,double *dout,double eout)
{
  double al,del,tau0,tau;
  double zeta,theta,z,ac,bc,cc,zx,a1,d1;
  const double C1=2.062648e5;
 
  if (ein == eout) {                   // input = output equinox?
    *aout = ain;
    *dout = din;
    return;
  }
  al  = ain*(M_PI/12.0);               // coordinates in radians
  del = din*(M_PI/180.0);
 
  tau0   = (ein-1900.0)/100.0;         // Newcomb: 
  tau    = (eout-1900.0)/100.0 - tau0;
  zeta   = (2304.25+1.396*tau0)*tau+.302*tau*tau+0.018*tau*tau*tau;
  z      = zeta+0.791*tau*tau+0.001*tau*tau*tau;
  theta  = (2004.682-0.853*tau0)*tau-0.426*tau*tau-0.042*tau*tau*tau;
  zeta  /= C1;
  z     /= C1;
  theta /= C1;
 
  ac = cos(del)*sin(al+zeta);
  bc = cos(theta)*cos(del)*cos(al+zeta)-sin(theta)*sin(del);
  cc = sin(theta)*cos(del)*cos(al+zeta)+cos(theta)*sin(del);
  d1 = asin(cc);
  zx = sqrt(1.0-cc*cc);
  double q = bc/zx;
  if      (q >  1.0) q =  1.0;
  else if (q < -1.0) q = -1.0;
  a1 = acos(q);
  if (ac/zx < 0.0) a1=2.0*M_PI-a1;
  a1 += z;
 
  *aout = a1*(12.0/M_PI);
  if (*aout <   0.0) *aout += 24.0;
  if (*aout >= 24.0) *aout -= 24.0;
  *dout = d1*(180.0/M_PI);
}

/* ---------------------------------------------------------------- */

void q_sort(int* zahl, int l, int r)   // quick-sort   
{
  int   x,h;
  int   i,j;
 
  i = l;
  j = r;
 
  if (j>i) {
    x = zahl[(i+j)/2];
    do {
      while (zahl[i] < x) i++;
      while (zahl[j] > x) j--;
      if (i<=j) {
        h         = zahl[i];
        zahl[i++] = zahl[j];
        zahl[j--] = h;
      }
    } while (!(i>j));
    q_sort(zahl,l,j);
    q_sort(zahl,i,r);
  }
}

/* ---------------------------------------------------------------- */

void f_sort(float* zahl, int l, int r)   // quick-sort   
{
  float x,h;
  int   i,j;
 
  i = l;
  j = r;
 
  if (j>i) {
    x = zahl[(i+j)/2];
    do {
      while (zahl[i] < x) i++;
      while (zahl[j] > x) j--;
      if (i<=j) {
        h         = zahl[i];
        zahl[i++] = zahl[j];
        zahl[j--] = h;
      }
    } while (!(i>j));
    f_sort(zahl,l,j);
    f_sort(zahl,i,r);
  }
}
 
/* ---------------------------------------------------------------- */

char* my_extend(char *s, char c, int l)
{ 
  size_t i = strlen(s);
  while (i < l) s[i++] = c;

  s[i] = '\0';

  return s;
}
 
/* --------------------------------------------------------------- */

char* cut_spaces(char* line)  // IDEA: stringByTrimmingCharactersInSet
{
  char *p1,*p2;

  p1 = line;
  p2 = line + strlen(line) -1;

  if (p2 < p1) return(line);

  if ((p1 == p2) && isspace(*p1)) { *p1 = '\0'; return(line); }

  while (isspace(*p1) && (p1 <  p2))               p1++;
  while (isspace(*p2) && (p2 >= p1)) { *p2 = '\0'; p2--; }

  if (p1 > line) { char *p0=line;      // copy to front of line
    while (*p1) {
      *p0 = *p1; p0++; p1++;
    }
    *p0 = '\0';
  }

  return line;
}

/* ---------------------------------------------------------------- */

int imin(int a,int b)
{
  if (a < b) return(a);
  return b;
}

/* ---------------------------------------------------------------- */

int imax(int a,int b)
{
  if (a > b) return a;
  return b;
}

/* ---------------------------------------------------------------- */

char* get_local_datestr(char* string,time_t ut) // NOTE: min.len=32
{
  struct tm *now,res;

  now = localtime_r(&ut,&res);
  strftime(string,32,"%b %d %H:%M:%S",now); // strlen(string)=15
  return string;
}

/* ---------------------------------------------------------------- */

char* get_local_timestr(char* string,time_t ut)
{
  struct tm *now,res;

  now = localtime_r(&ut,&res);
  strftime(string,16,"%H:%M:%S",now);  // min.length of 'string'
  return string;
}

/* ---------------------------------------------------------------- */

char* get_ut_timestr(char* string,time_t ut)
{
  struct tm *now,res;

  now = gmtime_r(&ut,&res);
  strftime(string,16,"%H:%M:%S",now);  // min.length of 'string'
  return string;
}

/* ---------------------------------------------------------------- */

double mag2flux(double mag,double cfzm)
{
  return(cfzm/pow(10.,0.4*mag));
}

/* --- */

double flux2peak(double flux,double fwhm,double pixscale)
{
  return(flux*(pixscale*pixscale)/(fwhm*fwhm*1.133));
}

/* --- */

double peak2flux(double peak,double fwhm)
{
  return(1.133*peak*fwhm*fwhm);
}

/* --- */

double flux2mag(double flux,double cfzm)
{
  return(2.5*log10(cfzm/flux));
}

/* ---------------------------------------------------------------- */

int is_bigendian(void)
{
  union { int l; char c[sizeof(int)]; } u;
  
  u.l = 1;
  return((u.c[sizeof(int)-1] == 1));
}

/* ---------------------------------------------------------------- */

double my_round(double x,int d)
{
  int i;
  
  for (i=0; i<d; i++) x *= 10.0;
  x = floor(x+0.5);
  for (i=0; i<d; i++) x /= 10.0;

  return(x);
}
  
/* ---------------------------------------------------------------- */

double modulo(double x,double min,double max)
{
  if (max <= min) return x;
  
  while (x >= max) x -= (max-min);
  while (x <  min) x += (max-min);
  
  return x;
}

/* ---------------------------------------------------------------- */

double modneg(double x,double max)
{
  while (x >   max/2.0) x -= max;
  while (x <= -max/2.0) x += max;
  
  return x;
}

/* ---------------------------------------------------------------- */

double modtop(double x,double max)
{
  if (x > max) {
    x = max - modulo(x,0.0,max);
  } else
  if (x < -max) {
    x = -max + modulo(fabs(x),0.0,max);
  }
  return x;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */


